#
# Copyright (C) 2011 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Looks after the logs for the session manager. Reading and Writing.
"""

import os
import sys
import logging

# We tried using json, but it wasn't good
# for config files, yaml used instead.
import yaml

from datetime import datetime
from labsession.config import CONFIG
LOGS = CONFIG.get('logs', '/tmp/')

def equal_to(first, second):
    """Sepcial equality checking"""
    if isinstance(first, list):
        return second in first
    else:
        return second == first


class SessionLogs(object):
   """Manages a directory of logs, allowing you to run reports on them.

   Usage:

   mylogs = SessionLogs(path='/var/log/session', date=datetime.now(),
       hostname=['david', 'golith'])

   for log in mylogs:
       print log.started

   Filename Based Limits:

     date     - Date of log creation is exactly equal to.*
     datelt   - Date of log creation is incluseivly less than.
     dategt   - Date of log creation is incluseivly greater than.
     time     - Time of log creation is exactly equal to.*
     timelt   - Time of log creation is incluseivly less than.
     timegt   - Time of log creation is incluseivly greater than.
     hostname - Hostname of session is exactly equal to.*

   * All equality checks can contain either a single item, or a list of items
     to check. Each one will be checked for equality.

   """
   def __init__(self, path=LOGS, **limits):
       self.path = path
       self.lims = limits

   def __iter__(self):
       for fname in os.listdir(self.path):
           if '.log' not in fname:
               continue
           (name, ext) = fname.split('.')[0:2]
           (hostname, year, month, day, hour, minute) = name.split('-')
           # All these limit checks can be done before opening the file.
           if self.lims:
               stamp = datetime(year, month, day, hour, minute)
               if not self.check_limits(hostname, stamp):
                   continue
           yield SessionLog(filename=os.path.join(self.path, fname))

   def check_limits(self, hostname, stamp):
       """Internal method for checking logs are within the limits"""
       date = stamp.date()
       time = stamp.time()
       l = self.lims
       # Date is equal to
       if l.has_key('date') and equal_to(l['date'], date):
           return False
       if l.has_key('datelt') and l['datelt'] > date:
           return False
       if l.has_key('dategt') and l['dategt'] < date:
           return False
       if l.has_key('time') and equal_to(l['time'], time):
           return False
       if l.has_key('timelt') and l['timelt'] > time:
           return False
       if l.has_key('timegt') and l['timegt'] < time:
           return False
       if l.has_key('hostname') and equal_to(l['hostname'], hostname):
           return False
       return True

class SessionLog(object):
    """Manages a session Log. You can load an existing log or start a new one."""
    def __init__(self, filename=None, hostname=None, username=None):
        if filename:
            self.load(filename)
        else:
            self.new_log(hostname, username)

    def new_log(self, hostname, username):
        self.started = datetime.now()
        self.events = []
        self.uniques = {}
        self.c = { 'meta' : {
            'hostname': hostname,
            'username': username,
          },
          'unique-events': self.uniques,
          'events': self.events,
        }
        # We use minute resolution on the log file.
        time = "%d-%d" % (self.started.hour, self.started.minute)
        date = self.started.date().isoformat()
        filename = "%s-%s-%s.log" % (hostname, date, time)
        self.path = os.path.join(LOGS, filename)
        self.event('log-started', date=self.started)

    def event(self, message, tag=None, date=None, data=None):
        """Record a log message:

           tag:   use a tag for this log, it's uniquly tagged, further events
                  with the same tag will overwrite.
           date:  Don't use the current datetime, use this one instead.
           data:  Extra data to save for this log event.

        """
        date = date or datetime.now()
        log = data or {}
        log['date'] = date.isoformat()
        log['message'] = message
        if tag:
            log['tag'] = str(tag)
            self.uniques[str(tag)] = log
        self.events.append(log)
        self.save()

    def save(self):
        """We're saving the structure to disk"""
        with open(self.path, 'w') as fhl:
            try:
                fhl.write(yaml.dump(self.c))
            except TypeError:
                sys.stderr.write("Unable to write: %s" % str(self.c))

    def load(self, filename):
        """Loading a filename log"""
        self.path = filename
        with open(self.path, 'r') as fhl:
            self.c = yaml.load(fhl.read())
            self.uniques = self.c.get('unique-events')
            self.events = self.c.get('events')


    def ended(self):
        """End the logfile"""
        self.event('log-ended')


