#!/usr/bin/python

import os
import sys
import atexit
import signal
import psutil
import socket
import getpass
import logging

from gi.repository import Gtk
from datetime import datetime, timedelta

import pynotify
import appindicator
from commands import getoutput

from labsession.config import CONFIG, TESTING
from labsession.logs import SessionLog

SESSION = CONFIG.get('session', "gnome-session")
AGROUP = CONFIG.get('limit_by_group', {'*': '01:00:00'})
AUSERS = CONFIG.get('limit_by_user', {})
WARNINGS = CONFIG.get('warnings', {})

class Session(object):
    """Manages a user's session."""
    def __init__(self):
        self.started = False
        self.hostname = socket.gethostname()
        self.username = getpass.getuser()
        self.max_length = None
        self.warnings = []

    @property
    def pid(self, session=SESSION):
        """The PID for the gnome session"""
        for proc in psutil.process_iter():
            if proc.name == session:
                return proc.pid

    @property
    def groups(self):
        """Return a list of groups our user is a member of."""
        result = []
        line = getoutput('groups %s' % self.username)
        for group in line.split(':')[-1].split(' '):
            if group.strip():
                result.append( group )
        return result

    def get_time_limit(self):
        """Get the amount of time the user is allowed to be logged in."""
        # As well as getting the amount of time the user should be allowed
        # To stay logged in, we should also record _how_ we made that choice.
        (name, value) = self.check_access(AUSERS, [self.username])
        atype = 'user'
        if value == None:
            (name, value) = self.check_access(AGROUP, self.groups)
            atype = 'group'
        logging.warn("Using %s configuration '%s'" % (atype, name))
        if value:
            return (self.delta(value), atype, name)
        return (None, atype, name)

    def time(self, value):
        """Turns a string '01:02:22' into a time 1 hour, 2 mins and 22 secs"""
        result = datetime.now().time()
        if isinstance(value, str):
            if value.count(':') == 2:
                result = datetime.strptime(value, '%H:%M:%S').time()
            elif value.count(':') == 1:
                result = datetime.strptime(value, '%M:%S').time()
        return datetime.combine(datetime.today().date(), result)

    def delta(self, value):
        """Turns string '01:02:22' into 3742 delta seconds."""
        mod = [1, 60, 3600]
        if isinstance(value, str):
            parts = value.split(':')
            parts.reverse()
            result = 0
            for i in range(3):
                if len(parts) > i:
                    result += int(parts[i]) * mod[i]
            return timedelta(seconds=result)
        if isinstance(value, int):
            return timedelta(seconds=value)
        return None

    def check_access(self, access, checklist):
        """Return access based on the configuration."""
        for (name, value) in access.iteritems():
            if name in checklist or name == '*':
                return (name, value)
        return (None, None)

    def start(self):
        """Start logging the session."""
        self.log = SessionLog(hostname=self.hostname, username=self.username)
        self.log.event('login', tag='started', data={'session-pid':self.pid})

        atexit.register( self.end )
        self.started = datetime.now()

        (self.max_length, atype, name) = self.get_time_limit()
        for warning in WARNINGS:
            self.add_warning(warning, atype, name)
        self.start_indicator()

    def add_warning(self, warning, access, name):
        """Return a warning if it applies to this user"""
        # Allow us to control what warnings are seen by what users.
        checking = False
        checked = None
        for check in ('group', 'user'):
            if warning.has_key('if_%s' % check):
                checking = True
                # If the session was by group and the warning is for user
                # or visa-verse, then return. If the specified user doesn't match
                # Then also return. This warning will not be seen then.
                if access == check and name in warning['if_%s' % check]:
                    checked = True
        if checking and not checked:
            return
        for time in warning['times']:
            # Don't add already expired warnings.
            if not self.if_warning(warning, self.delta(time)):
                logging.warn("Adding warning: %s: %s" % (warning['message'], time) )
                self.warnings.append((self.delta(time), warning))

    def end(self, endtype='died'):
        """End the session, but don't force it's end."""
        if self.started:
            self.started = False
            self.log.event(endtype, tag='end')
            self.quit()

    def quit(self):
        """Close the gtk main loop"""
        Gtk.main_quit()
        if not TESTING:
            os.kill(self.pid, signal.SIGKILL)
        else:
            sys.stderr.write("Would have killed %d\n" % self.pid )

    def event(self, t, icon):
        """Log any event, internal methiod."""
        self.item.set_text(t)
        self.ind.set_attention_icon(icon)

    def if_warning(self, warning, wdelta, delta=None):
        """Returns true if the warning should be shown"""
        ctype = warning.get('from', 'session-end')
        delta = delta or timedelta(seconds=0)
        if ctype == 'session-end' and self.max_length:
            return delta > self.max_length - wdelta
        elif ctype == 'session-start':
            return delta > wdelta
        else: # From a sepcific time
            return datetime.now() + wdelta > self.time(ctype)

    def update_timeitem(self, datum=None):
        """Updates the menu item that reports the session time"""
        delta = datetime.now() - self.log.started
        if self.max_length and delta > self.max_length:
            return self.end('forced')
        elif self.max_length:
            rdelta = self.max_length - delta
            self.timeitem.set_label("Time Left: %s" % str(rdelta).split('.')[0])
        else:
            self.timeitem.set_label("No Time Limit")

        # Should we show any warnings?
        for i in range(len(self.warnings)):
            (wdelta, warning) = self.warnings[i]
            if self.if_warning(warning, wdelta, delta):
                self.send_warning(wdelta, warning)
                # Don't show this warning again
                self.warnings.pop(i)
                # Only show this warning and no other
                break

        # In about a second, redo the update.
        Gtk.timeout_add(1000, self.update_timeitem, None)

    def send_warning(self, delta, warning):
        """Sends a warning to the system"""
        logging.warn("Hit %s warning" % str(warning))
        if warning.get('from', '') == 'session-end':
            # Change the state of the indicator to show we have to logoff.
            self.ind.set_status(appindicator.STATUS_ATTENTION)
        # Prepare all the data for the warning
        rmins = -1
        if self.max_length:
            rmins = self.max_length - delta
        jar = { 'time': int(delta.seconds / 60),
            'remains': int(rmins.seconds / 60) }
        mesg = warning.get('message', 'Unknown Warning') % jar
        desc = warning.get('description', '') % jar
        icon = warning.get('icon', 'gtk-info')
        # Send a notification to the system notifier.
        pynotify.Notification(mesg, desc, icon).show()

    def end_clicked(self, widget=None):
        """User clicked quit in the menu."""
        self.end('logoff')

    def start_indicator(self):
        """Start the indicator."""
        status = appindicator.CATEGORY_APPLICATION_STATUS
        self.ind = appindicator.Indicator('session-manager', "gtk-close", status)
        self.ind.set_icon('application-community')
        self.ind.set_attention_icon('dialog-warning')
        self.ind.set_status(appindicator.STATUS_ACTIVE)

        # create a menu
        self.menu = Gtk.Menu()
        if TESTING:
            title = Gtk.MenuItem("Testing Session...")
            self.menu.append(title)
            title.show()
            separator = Gtk.MenuItem()
            separator.show()
            self.menu.append(separator)

        self.timeitem = Gtk.MenuItem("...Please Wait")
        self.menu.append(self.timeitem)
        self.timeitem.show()

        separator = Gtk.MenuItem()
        separator.show()
        self.menu.append(separator)

        exit = Gtk.MenuItem("Exit")
        self.menu.append(exit)
        exit.connect('activate', self.end_clicked)
        exit.show()

        self.ind.set_menu(self.menu)
        # Time so gtk main loop has a chance to load before it's killed.
        Gtk.timeout_add(1000, self.update_timeitem, None)

